/*
 * Board.java
 *
 * Created on 6. November 2007, 18:46
 */

/*
 The contents of this file are subject to the
 Mozilla Public License Version 1.1 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied.
 See the License for the specific language governing rights
 and limitations under the License.

 The Initial Developer of the Original Code is
 Georg-Johann Lay (GeorgJohann@web.de).

 Portions created by the Initial Developer are
 Copyright (C) 2007 Georg-Johann Lay. All Rights Reserved.

 Alternatively, the contents of this file may be used under the terms
 of either the GNU General Public License Version 3 or later (the "GPL"),
 or the GNU Lesser General Public License Version 3 or later (the "LGPL"),
 in which case the provisions of the GPL or the LGPL are applicable instead
 of those above.

 If you wish to allow use of your version of this file only
 under the terms of either the GPL or the LGPL, and not to allow others to
 use your version of this file under the terms of the MPL, indicate your
 decision by deleting the provisions above and replace them with the notice
 and other provisions required by the LGPL or the GPL. If you do not delete
 the provisions above, a recipient may use your version of this file under
 the terms of any one of the MPL, the GPL or the LGPL.
*/

package de.gjlay.chess;

/**
 * A chess board
 * @author Georg-Johann Lay
 */
class Board
        implements BOARD_CONST, CoDec
{

    /**
     * Encodes arrangement of pieces in the base line
     */
    private static final int[] BASE  =
    {
        ROOK, KNIGHT, BISHOP, QUEEN, KING, BISHOP, KNIGHT, ROOK
    };

    /**
     * Array of 8 &times; 8 representing the very board. Each entry is {@link #EMPTY} or 
     * one of {@link #PAWN}, {@link #KNIGHT}, {@link #BISHOP}, {@link #ROOK}, {@link #QUEEN}
     * or {@link #KING} ored with the owner's color,
     * i.e. {@link #BLACK} (1) or {@link #WHITE} (0). Therefore, a field will hold one of 
     * 13 possible values.
     */
    int[][] p;
    /**
     * Was last move a double step of a pawn? Needed for e.p.
     */
    int lastPawn2x = -1;
    /**
     * Array with 2 entries holding the kings' position(s) or <CODE>null</CODE> if
     * the respective king is missing (may occur when compising a position).
     */
    Field[] king;
    /**
     * Mover that is to move next, one of {@link #BLACK} or {@link #WHITE}.
     */
    int mover;
    /**
     * Start position (needs to be) saved?
     */
    boolean startSaved = true;
    /**
     * Lowest 4 bits encode if or if not castling is allowed. 
     * This represents the "global" to do the respective castling and is not affected
     * by {@link #mover} or if castling is allowed in the concrete position.
     */
    int mayCastle;
    /**
     * Start position with check?
     */
    boolean startsWithCheck;
    /**
     * Can player do a move? (Mind checkmate, patt). Used in cursor drawing.
     */
    boolean canMove;

    /**
     * The history which is needed to {@link #undo()} and {@link #redo()} moves.
     */
    History h;

    /**
     * Get piece id without color
     * @param x column
     * @param y row
     * @return id (without color) of piece, i.e. one of {@link #EMPTY}, 
     * {@link #PAWN}, {@link #KNIGHT}, {@link #BISHOP}, {@link #ROOK}, {@link #QUEEN}
     * or {@link #KING}.
     */
    int p (int x, int y)
    {
        return p[x][y] & 0xfffffffe;
    }

    /**
     * Get color of piece at field
     * @param x column
     * @param y row
     * @return color, one of {@link #BLACK} or {@link #WHITE}.
     */
    int c (int x, int y)
    {
        return p[x][y] & 1;
    }

    /**
     * Masks for 0-0
     */
    static final int[] M_0_0   = { (1<<0), (1<<1) };
    /**
     * Masks for 0-0-0
     */
    static final int[] M_0_0_0 = { (1<<2), (1<<3) };

    /**
     * Creates a new instance of Board
     */
    Board ()
    {
        p    = new int[8][8];
        king = new Field[2];

        h = new History ();
    }

    /**
     * Construct a board with a given history
     * @param moves array of moves like <CODE>"e2e4"</CODE>, <CODE>"e8g8"</CODE> or <CODE>"a7b8D"</CODE>.
     */
    Board (String[] moves)
    {
        this();
        reset();
        
        setHistory (moves);
    }
    
    /**
     * Construct a board with given position and history
     * @param moves array of moves like <CODE>"e2e4"</CODE>, <CODE>"e8g8"</CODE> or <CODE>"a7b8D"</CODE>.
     */
    Board (Exercise ex)
    {
        this();
        ex.fillBoard (this);
    }

    /**
     * Clear board. No pieces will be there after calling.
     */
    void clear()
    {
        for (int x = 0; x < 8; x++)
        for (int y = 0; y < 8; y++)
            p[x][y] = EMPTY;

        king[WHITE] = king[BLACK] = null;

        startSaved = false;
        startsWithCheck = false;
        canMove = true;
    }

    /**
     * Set starting position for a new game
     */
    void reset()
    {
        for (int x = 0; x < 8; x++)
        for (int y = 0; y < 8; y++)
            p[x][y] = getResetPiece (x, y);

        mayCastle = M_0_0[WHITE] | M_0_0_0[WHITE]
              | M_0_0[BLACK] | M_0_0_0[BLACK];

        setKing (4, 0);
        setKing (4, 7);

        mover = WHITE;

        h.clear();
        startSaved = false;
        startsWithCheck = false;
        canMove = true;
    }
    
    void setHistory (String[] moves)
    {
        h.clear();
        
        for (int i=1; i < moves.length; i++)
        {
            Move move = Move.FromString (this, moves[i]);
            apply (move);
            move.isCheck = isControlled (1-mover, king[mover]);
        }

        h.undoAll (this);
    }            

    void setKing (int x, int y)
    {
        if (KING == p(x,y))
            king[c(x,y)] = Field.Get (x, y);
    }

    /**
     * Get id of piece in start position
     * @param x column
     * @param y row
     * @return id of piece
     */
    int getResetPiece (int x, int y)
    {
        switch (y)
        {
            case 7: return BLACK | BASE[x];
            case 6: return BLACK | PAWN;
            case 1: return WHITE | PAWN;
            case 0: return WHITE | BASE[x];
        }
        return EMPTY;
    }

    char toChar (int x, int y)
    {
        int pp = p[x][y];

        return (WHITE == (pp & 1))
            ? Character.toUpperCase (PIECES_CHARS.charAt(pp >> 1))
            : Character.toLowerCase (PIECES_CHARS.charAt(pp >> 1));
    }

    /**
     * 
     * Coordinates in bounds?
     * @param x column
     * @param y row
     * @return true iff (x,y) in [0,7]&times;[0,7].
     */
    static boolean Ok (int x, int y)
    {
        return x >= 0 && x < 8
            && y >= 0 && y < 8;
    }

//#if CONSOLE
//#     public String toString()
//#     {
//#         String line = "";
//# 
//#         for (int y=0; y < 8; y++)
//#         {
//#             for (int x=0; x < 8; x++)
//#                 line += toChar (x, 7-y);
//# 
//#             line += "\n";
//#         }
//# 
//#         return line;
//#     }
//#endif

//#if CONSOLE                
//#         void dump ()
//#         {
//#             String str, in3 = "            ", in2 = "        ";
//#             
//#             System.out.println (in2+"new Exercise (\"\", ");
//#             System.out.println (in3+mover+",");
//#             
//#             for (int c = WHITE; c <= BLACK; c++)
//#             {
//#                 str = in3 + ((mayCastle & M_0_0[c]) != 0) + ", "
//#                           + ((mayCastle & M_0_0_0[c]) != 0) + ",";
//#                 System.out.println (str);
//#                 System.out.println (in3+getPieceDump(c)+",");
//#             }
//#             System.out.println (in3+lastPawn2x+",");
//#             
//#             h.dump (History.DUMP_STRING, true);
//#             
//#             System.out.println (")");
//#         }
//#         
//#         /// Get String representation for all pieces of color
//#         /// beginning with most important pieces
//#         private String getPieceDump (int color)
//#         {
//#             String str = "";
//#             for (int pi = KING; pi >= PAWN; pi -= 2)
//#             {
//#                 int pc = pi | color;
//#                 
//#                 for (int x=0; x<8; x++)
//#                 for (int y=0; y<8; y++)
//#                     if (pc == p[x][y])
//#                         str += " " + getString (x,y);
//#             }
//#             
//#             return "\"" + str.substring(1) + "\"";
//#         }
//#endif

    private final static int DXY[][] =
    {
        { KNIGHT, 2, 1, 1},  { KNIGHT, 1, 2, 1},
        { BISHOP, 1, 1, 7},
        { ROOK,   1, 0, 7},
        { QUEEN,  1, 1, 7},  { QUEEN,  1, 0, 7},
        { KING,   1, 1, 1},  { KING,   1, 0, 1}
    };

    /**
     * Is a field controlley by a player?
     * @param color color of player
     * @param f the field to test
     * @return true if controlled (i.e. attacked)
     */
    boolean isControlled (int color, Field f)
    {
        for (int y=0; y < 8; y++)
            for (int x=0; x < 8; x++)
                if (p(x,y) != EMPTY
                    && c(x,y) == color
                    && getControlledFields (x, y).contains (f))
                        return true;

        return false;
    }

    /**
     * Get list of controlled fields
     * @param x column
     * @param y row
     * @return all fields controlled by piece at (x,y)
     */
    private Fields getControlledFields (int x, int y)
    {
        Fields controlledFields = new Fields (this);

        int id;

        switch (id = p(x, y))
        {
            case EMPTY:
                    break;

            case KNIGHT:
            case BISHOP:
            case ROOK:
            case QUEEN:
            case KING:
                for (int i=0; i < DXY.length; i++)
                {
                    int[] args = DXY[i];

                    if (id == args[0])
                        controlledFields.addRay (x, y, args[1], args[2], args[3]);
                }

                break;

            case PAWN:
            {
                int dy = (WHITE == c(x,y)) ? 1 : -1;

                controlledFields.add (x+1, y+dy);
                controlledFields.add (x-1, y+dy);

                break;
            }
        }

        return controlledFields;
    }

    String getString (int x, int y)
    {
        int pp = p(x,y);
        return ((pp == EMPTY)
                ? ""
                : ((pp == PAWN)
                    ? ""
                    : "" + Character.toUpperCase (toChar(x,y))))
                + Field.GetString (x, y);
    }

    /**
     * Can player to move actually do a move?
     * @return true if there is at least one legal move, false otherwise (checkmate, patt)
     */
    boolean canMove ()
    {
        for (int y=0; y < 8; y++)
        for (int x=0; x < 8; x++)
            if (canMove (x,y))
                return true;

        return false;
    }

    /**
     * Is piece allowed to move?
     * @param x column
     * @param y row
     * @return true if a piece at (x,y) is allowed to move, false otherwise
     */
    private boolean canMove (int x, int y)
    {
        if (EMPTY == p[x][y]
                || (p[x][y] & 1) != mover)
            return false;

        Moves moves = new Moves();
        Fields movetoFields = getMovetoFields (x, y);

        for (int i=0; i < movetoFields.size(); i++)
        {
            Field f_to   = (Field) movetoFields.elementAt (i);
            Field f_from = Field.Get (x, y);

            Move move = testAddMove (moves, f_from, f_to, QUEEN | mover);

            if (null != move)
                return true;
        }

        return false;
    }

    /**
     * Get all possible moves
     * @param x column
     * @param y row
     * @return all moves that piece at (x,y) can do
     */
    Moves getMoves (int x, int y)
    {
        Moves moves = new Moves();

        if (EMPTY == p(x,y))
            return moves;

        int color = c(x,y);

        if (mover != color)
            return moves;

        Fields movetoFields = getMovetoFields (x, y);

        for (int i=0; i < movetoFields.size(); i++)
        {
            Field f_to   = (Field) movetoFields.elementAt (i);
            Field f_from = Field.Get (x, y);

            Move move = testAddMove (moves, f_from, f_to, QUEEN | color);

            if (null == move || !move.isPromote)
                continue;

            testAddMove (moves, f_from, f_to, ROOK   | color);
            testAddMove (moves, f_from, f_to, BISHOP | color);
            testAddMove (moves, f_from, f_to, KNIGHT | color);
        }

        return moves;
    }

    // mover's king attacked?
    /**
     * Check for player of given color?
     * @param mover color of player
     * @return return true if player's king is in check
     */
    boolean isCheck (int mover)
    {
        if (null == king[mover])
            return false;
        return isControlled (1-mover, king[mover]);
    }

    /**
     * Check?
     * @return true if mover is in check, false otherwise
     */
    boolean isCheck ()
    {
        if (0 != h.am)
        {
            Move move = h.getElement (h.am-1);
            return move.isCheck;
        }
        else
        {
            return startsWithCheck;
        }
    }

    /**
     * unternal use
     */
    private Move testAddMove (Moves moves, Field f_from, Field f_to, int morphTo)
    {
        Move move = new Move (this, f_from, f_to, morphTo);

        // try
        move.apply (this);

        int color = 1 & morphTo;
        boolean moveOk = !isCheck (color);

        if (moveOk)
        {
            move.isCheck = isCheck (1-color);
            moves.addElement (move);
        }

        move.undo (this);

        return (moveOk) ? move : null;
    }

    /**
     * Collect all fields that can be reached from the piece at (x,y).
     * (x,y) may be {@link #EMPTY}.
     * Does not mind wether or not the resulting move leads to check .
     * @param x column
     * @param y row
     * @return fields piece at (x,y) can reach
     */
    private Fields getMovetoFields (int x, int y)
    {
        Fields movetoFields = getControlledFields (x, y);

        int id = p(x,y);

        if (PAWN == id)
            treatPawnMoveFields (movetoFields, x, y);

        if (KING == id)
            treatKingCastlingFields (movetoFields, x, y);

        int color = c(x,y);

        for (int i = movetoFields.size()-1; i >= 0; i--)
        {
            Field field = (Field) movetoFields.elementAt(i);

            // no moves to occupied fields
            if (p(field.x, field.y) != EMPTY
                    && color == c(field.x, field.y))
                        movetoFields.removeElementAt (i);
        }

        return movetoFields;
    }

    /**
     * Add castling destinations (if any, {@link #KING}'s field only) to <code>moveto</code>.
     * @param moveto moves to add to
     * @param x 
     * @param y 
     */
    private void treatKingCastlingFields (Fields moveto, int x, int y)
    {
        // castling

        int color = c(x,y);

        if (! king[color].equals (Field.Get (x, y)))
            throw new NullPointerException ("King info corrupt");

        if (! king[color].equals (Field.Get (4, (WHITE == color) ? 0 : 7)))
            return;

        // K in check

        if (isControlled (1-color, Field.Get (x,y)))
            return;

        for (int dx = -1; dx <= 1; dx += 2)
        {
            int xRook = (dx == -1) ? 0 : 7;

            int mask = (-1 == dx) ? M_0_0_0[color] : M_0_0[color];
            if (0 == (mayCastle & mask)
                || EMPTY != p(4+dx, y)
                || EMPTY != p(4+2*dx, y)
                || EMPTY != p(xRook-dx, y)
                || isControlled (1-color, Field.Get (4+dx,y)))
                continue;

            moveto.add (4+2*dx, y);
        }
    }


    /**
     * Add pawn moves' destinations to <code>moveto</code>.
     * Handles double step, en passant (e.p.) etc.
     * @param moveto
     * @param x
     * @param y
     */
    private void treatPawnMoveFields (Fields moveto, int x, int y)
    {
        int color = c(x, y);
        int dy = (WHITE == color) ? 1 : -1;
        int y0 = (WHITE == color) ? 1 : 6;

        // en passant
        for (int i = moveto.size()-1; i >= 0; i--)
        {
            Field field = (Field) moveto.elementAt(i);
            if (EMPTY != p (field.x, field.y))
                // move to attacked field is allowed if not empty
                continue;

            if (y != y0+3*dy
                || field.x != lastPawn2x
                || PAWN != p (field.x, y))
                    moveto.removeElementAt (i);
        }

        if (EMPTY == p(x, y+dy))
        {
            // single step
            moveto.add (x, y+dy);

            // double step
            if (y == y0 && EMPTY == p(x, y+2*dy))
                moveto.add (x, y+2*dy);
        }
    }

    /** Undo last move (if any)
     * @return The move that is undone or <code>null</code>.
     * @see History#undo(Board)
     */
    Move undo ()
    {
        Move m = h.undo (this);
        canMove = canMove();
        return m;
    }

    /** Redo move (if any)
     * @return The move that is redone or <code>null</code>.
     * @see History#redo(Board)
     */
    Move redo ()
    {
        Move m = h.redo (this);
        canMove = canMove();
        return m;
    }

    /** Apply (i.e. do) move <code>m</code>
     * @param m The move that is to be done. <code>m</code> will be the last move in the history.
     * @see History#apply(Board,Move)
     */
    void apply (Move m)
    {
        h.apply (this, m);
        canMove = canMove();
    }

    Move removeLastMove ()
    {
        Move m = undo ();

        if (null != m)
            h.setSize (h.size()-1);

        return m;
    }

    /**
     * Load board and history form RecordStore
     * @param name identifier
     * @return true on success, otherwise the board will have been resetted
     * @see Lover
     */
    boolean load (String name)
    {
        byte[] code;

        code = Lover.Load (name+".board");

        boolean found = null != code;

        if (!found)
            reset();
        else
            decode (code, null);

        startsWithCheck = isCheck (mover);
        canMove = canMove();

//        Drawer.Load (name, mover);
        ///////////////////////////////////////
        code = Lover.Load (name+".history");

        if (null == code)
            h.clear();
        else
            h.decode (code, this);

        return found;
    }

    void save (String name)
    {
        if (!startSaved)
        {
            Lover.Save (name+".board", this);
            startSaved = true;
        }

//        Drawer.Save (name);
        Lover.Save (name+".history", h);
    }

    /**
     * 
     * @return size
     */
    public int getEncodeSize()
    {
        return   1 // lastPawn2x
                +1 // mayCastle
                +1 // mover
                +8*8 // board
        ;
    }

    /**
     * 
     * @return code
     */
    public byte[] encode ()
    {
        byte[] code = new byte[getEncodeSize()];

        code[0] = (byte) lastPawn2x;
        code[1] = (byte) mayCastle;
        code[2] = (byte) mover;

        int i=3;

        for (int x=0; x<8; x++)
        for (int y=0; y<8; y++)
            code[i++] = (byte) p[x][y];

        return code;
    }

    /**
     * 
     * @param code 
     * @param _unused_ 
     */
    public void decode (byte[] code, Object _unused_)
    {
        lastPawn2x = code[0];
        mayCastle      = code[1];
        mover      = code[2];

        int i=3;

        for (int x=0; x<8; x++)
        for (int y=0; y<8; y++)
        {
            int pp = p[x][y] = code[i++];

            if (p(x,y) == Board.KING)   king[pp & 1] = Field.Get (x,y);
        }
    }
}
