/*
 * PlayCanvas.java
 *
 * Created on 4. November 2007, 20:51
 */

/*
 The contents of this file are subject to the
 Mozilla Public License Version 1.1 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at

 http://www.mozilla.org/MPL/

 Software distributed under the License is distributed on an "AS IS" basis,
 WITHOUT WARRANTY OF ANY KIND, either express or implied.
 See the License for the specific language governing rights
 and limitations under the License.

 The Initial Developer of the Original Code is
 Georg-Johann Lay (GeorgJohann@web.de).

 Portions created by the Initial Developer are
 Copyright (C) 2007 Georg-Johann Lay. All Rights Reserved.

 Alternatively, the contents of this file may be used under the terms
 of either the GNU General Public License Version 3 or later (the "GPL"),
 or the GNU Lesser General Public License Version 3 or later (the "LGPL"),
 in which case the provisions of the GPL or the LGPL are applicable instead
 of those above.

 If you wish to allow use of your version of this file only
 under the terms of either the GPL or the LGPL, and not to allow others to
 use your version of this file under the terms of the MPL, indicate your
 decision by deleting the provisions above and replace them with the notice
 and other provisions required by the LGPL or the GPL. If you do not delete
 the provisions above, a recipient may use your version of this file under
 the terms of any one of the MPL, the GPL or the LGPL.
*/

package de.gjlay.chess;

import de.gjlay.mobile.localize.*;
import de.gjlay.mobile.ui.*;
import java.io.InputStream;
import java.util.Hashtable;
import javax.microedition.lcdui.*;
import javax.microedition.midlet.MIDlet;

/**
 *
 * @author Georg-Johann Lay
 */
final class PlayCanvas
    extends ChessCanvas
    implements SelectListener, MessageListener, LanguageListener
{
    private static final String LOVER_ID = "last";
    
    private static final int SEL_FROM  = 0;
    private static final int SEL_TO    = 1;
    private static final int SEL_MORPH = 2;

    private static final int ID_LANG  = 1;
    private static final int ID_EXIT  = 2;
    private static final int ID_RESET = 3;
    private static final int ID_TURN  = 4;
    private static final int ID_LOAD  = 5;
    private static final int ID_SAVE  = 6;
    private static final int ID_POSITION  =  7;
    private static final int ID_HELP      =  8;
    private static final int ID_BACKLIGHT =  9;
    private static final int ID_FAMOUS    = 10;
    private static final int ID_UNDO_ALL  = 11;
    private static final int ID_EXERCISE  = 12;

    private static final int ID_LANG_DEFAULT = 0;
    private static final int ID_DE = 1;
    private static final int ID_EN = 2;
    private static final int ID_FR = 3;

    // selectionDone
    private static final int TAG_MAIN_MENU = 0;
    private static final int TAG_LANG_MENU = 1;
    private static final int TAG_FAMOUS    = 2;
    private static final int TAG_EXERCISE  = 3;

    // messageReturned
    private static final int TAG_RESET = 1;
    private static final int TAG_HELP  = 2;
    private static PlayCanvas instance;

    private SelectorLabel[] mainLabels;
    private SelectorLabel[] langLabels;

    static SelectorLabel lightLabel;

    private Board b;

    private int px, py;

    /** Will be one of {@link #SEL_FROM}, {@link #SEL_TO}, {@link #SEL_MORPH}.
     */
    private int stage = SEL_FROM;
    private int mover;

    /** {@link #move} will be from this {@link Field}. */
    private Field f_from;

    /** {@link #move} will be to that {@link Field}. */
    private Field f_to;

    private Moves moves = null;

    private Move move;
    private Localizer localizer;

    /** Indicate which piece to move next when redoing
     * moves. This is nice when stepping through famous games
     * so we don't get a hint of what comes next...
     */
    private boolean peekNextMove = true;

    private int mainIndex = 0;

    private int promoteIndex = 0;

    PlayCanvas (MIDlet midlet, Localizer localizer)
    {
        super (midlet);
        this.localizer = localizer;
        init();
    }

    static PlayCanvas GetInstance (MIDlet midlet, Localizer localizer)
    {
        return
            instance = new PlayCanvas (midlet, localizer);
    }

    static PlayCanvas GetInstance ()
    {
        return instance;
    }

    private void init()
    {
        //languageChanged ();
        localizer.addLanguageListener (this);
    }

    private void initCursor()
    {
        Move m;
        
        mover = b.mover;

        if (null != (m = undo (false)))
        {
            m = redo();
            px = m.f_to.x;
            py = m.f_to.y;
        }
        else if (peekNextMove && null != (m = redo()))
        {
            m = undo (false);
            px = m.f_from.x;
            py = m.f_from.y;
        }
        else
        {
            px = 0;
            py = 0;
        }
    }
/*
    final static void debug (String msg)
    {
        System.err.println (msg);
    }
*/
    private Move undo (boolean doSave)
    {
        Move m = b.undo();
        if (doSave)
        {
            b.save (LOVER_ID);
            save (LOVER_ID);
        }
        return m;
    }

    private Move redo()
    {
        Move m = b.redo();
        b.save (LOVER_ID);
        save (LOVER_ID);
        return m;
    }

    private void apply (Move m, boolean doSave)
    {
        b.apply (m);
        if (doSave)
        {
            b.save (LOVER_ID);
            save (LOVER_ID);
        }
        peekNextMove = true;
    }

    public void keyPressed (int code)
    {
        cursorCount = 0;
        boolean done = true;

        switch (code)
        {
            default:
                done = false;
                break;

            case KEY_STAR:
            case KEY_POUND:
                handleUndoRedo (code);
                break;

            case KEY_NUM1:
                addMode (M_FIX, mover);
                break;

            case KEY_NUM2:
                addMode (M_TURN2, mover);
                break;

            case KEY_NUM3:
                addMode (M_FIX2, mover);
                break;

            case KEY_NUM4:
                addMode (M_FLIP2, mover);
                break;

            case KEY_NUM7:
                ChessCanvas.BacklightOn ^= true;
                break;

            case KEY_NUM8:
                showLetters = !showLetters;
                break;

            case KEY_NUM9:
                mainIndex = 0;
                showMainMenu();//Message.Show (this, TAG_RESET, Message.CONFIRM, L._("T_ASK_BOARD_RESET"));
                return;

            case KEY_NUM0:
            case KEY_NUM5:
            case KEY_NUM6:
                break;
        }

        if (done)
        {
            save (LOVER_ID);
            repaint();
            return;
        }

        int game = getGameAction (code);

        if (stage == SEL_MORPH)
            handleMorph (game);
        else
            handlePlay (game);

//        debug ("M "+b.getMoves (px, py)+"\n");

        save (LOVER_ID);
        repaint();
    }

    private void handlePlay (int game)
    {
        boolean done = true;

        int dd = turn ? -1 : 1;

        switch (game)
        {
            default:               return;
            case FIRE:             done = false;            break;
            case UP:               py += dd;                break;
            case DOWN:             py -= dd;                break;
            case RIGHT:            px += dd;                break;
            case LEFT:             px -= dd;                break;
        }

        if (done)
        {
            px = Math.min (7, Math.max(0, px));
            py = Math.min (7, Math.max(0, py));
            return;
        }

        // FIRE
        switch (stage)
        {
            case SEL_FROM:

                moves = b.getMoves (px, py);
                if (moves.isEmpty ())
                {
                    moves = null;
                    return;
                }

                stage = SEL_TO;
                f_from = Field.Get (px, py);

                //debug ("Moves["+stage+"] = " + moves);
                break;

            case SEL_TO:
                f_to = Field.Get (px, py);

                if (f_to.equals (f_from))
                {
                    stage = SEL_FROM;
                    break;
                }

                Moves movesF = moves.filter (f_from, f_to);

                if (movesF.size() == 0)
                    break;
                else
                    moves = movesF;

                if (moves.size() == 1)
                {
                    move = (Move) moves.firstElement();
                    apply (move, true);
                    stage = SEL_FROM;
                    moves = null;
                    mover = b.mover;
                    flushMode (mover);
                    break;
                }

                if (moves.size() > 1)
                {
                    move = (Move) moves.firstElement();
                    apply (move, false);
                    promoteIndex = 0;
                    stage = SEL_MORPH;
                }
                break;
        }
    }

    private void handleMorph (int game)
    {
        switch (game)
        {
            case UP:
            case RIGHT:
                promoteIndex += 2;
            case DOWN:
            case LEFT:
                promoteIndex --;

                move = (Move) moves.elementAt (promoteIndex = (4+promoteIndex) % 4);
                undo (false);
                apply (move, false);
                break;

            case FIRE:
                undo (false);
                apply (move, true);
                stage = SEL_FROM;
                mover = b.mover;
                flushMode (mover);
                break;
        }
    }

    private void handleUndoRedo (int code)
    {
        Field f = null;
        Move  m;

        if (stage == SEL_MORPH)
        {
            m = b.removeLastMove();
            f = m.f_from;
        }
        else
        {
            if (KEY_POUND == code)
            {
                if (!peekNextMove)
                {
                    if ((m = redo ()) != null)
                        f = m.f_to;
                }
                else
                {
                    Move m1 = redo();
                    m = b.redo ();
                    
                    if (m != null)
                    {
                        f = m.f_from;
                        b.undo ();
                    }
                    else if (m1 != null)
                        f = m1.f_to;
                }
            }

            if (KEY_STAR == code)
                if ((m = undo (true)) != null)
                    f = m.f_from;
        }

        stage = SEL_FROM;
        if (f != null)
        {
            px = f.x;
            py = f.y;
            flushMode (b.mover);
        }

        mover = b.mover;
//        debug ("H="+b.h);
    }

    public void paint (Graphics g)
    {
        setDrawContext (g, b, mover);

        g.translate(- g.getTranslateX(), - g.getTranslateY());

        drawBackground (0x006000);

        g.translate (getWidth()/2-4*SIZE, getHeight()/2-4*SIZE);

        drawFrame ();
        drawBoard ();

        if (b.canMove || stage == SEL_MORPH)
            drawCursors ();
        drawPieces ();
    }

    protected void drawPieces ()
    {
        super.drawPieces ();
        // Schach ?
        if (b.isCheck())
            drawCheck (b.mover);
    }

    private void drawCursors ()
    {
        boolean high = (stage == SEL_TO
                    && px == f_from.x && py == f_from.y)
                    || stage == SEL_MORPH;
        drawCursor (high ? CURSOR_MARKBLINK : CURSOR_BLINK, px, py);

        if (!high && stage == SEL_TO)
            drawCursor (CURSOR_MARK, f_from.x, f_from.y);
    }

    static void Show (Board b)
    {
        instance.startGame (b);
    }

    private void startGame (Board b)
    {
        this.b = b;
        b.startSaved = false;
        b.canMove = b.canMove();
        b.save (LOVER_ID);
        save (LOVER_ID);
        
        initCursor();
        resumeGame();
    }

    static void Show (boolean newGame)
    {
        if (newGame)
            instance.startGame();
        else
            instance.resumeGame();
    }

    private void startGame()
    {
        b = new Board();
        b.load (LOVER_ID);
        load (LOVER_ID, b.mover);

        initCursor();
        resumeGame();
    }

    private void resumeGame()
    {
        load (LOVER_ID, mover);
        //Drawer.canvas = this;
        setFullScreenMode (true);
        Display.getDisplay (midlet).setCurrent (this);

        repaint();

//        debug (""+b);
    }

    void resetBoard()
    {
        b.reset();
        b.save (LOVER_ID);
        save (LOVER_ID);

        mover = b.mover;
        px = 4; //b.king[WHITE].x;
        py = 0; //b.king[WHITE].y;

        peekNextMove = true;

        stage = SEL_FROM;
    }

    private void showMainMenu()
    {
        for (int i=0; i < mainLabels.length; i++)
            if (mainLabels[i].equals(lightLabel))
                mainLabels[i] = GetLightLabel();

        Selector.Show (this, TAG_MAIN_MENU, L._("T_TITLE"), mainLabels);
        
        if (mainIndex >= 0)
            Selector.SetSelectedIndex (mainIndex);
    }

    private void showFamousMenu()
    {
        Selector.Show (this, TAG_FAMOUS, L._("T_TITLE"), Famous.LABELS);
    }

    private void showExerciseMenu()
    {
        Selector.Show (this, TAG_EXERCISE, L._("T_TITLE"), Exercise.LABELS);
    }

    private void showLanguageMenu()
    {
        Selector.Show (this, TAG_LANG_MENU, L._("T_TITLE"), langLabels);
    }

    public void selectionDone (int id, int tag, int index)
    {
        if (TAG_FAMOUS == tag 
                || TAG_EXERCISE == tag)
        {
            if (Selector.BACK == id)
            {
                showMainMenu();
                return;
            }
            
            peekNextMove = false;
        }
        
        if (TAG_FAMOUS == tag)
            Show (new Board (Famous.GAMES[id]));
        
        if (TAG_EXERCISE == tag)
            Show (new Board (Exercise.EXERCISES[id]));

        if (TAG_MAIN_MENU == tag)
        {
            mainIndex = index;
            handleMainMenu (id);
        }

        if (TAG_LANG_MENU == tag)
            handleLanguageMenu (id);
    }

    private void handleLanguageMenu (int id)
    {
        String locale = null;

        switch (id)
        {
            case ID_LANG_DEFAULT:
                locale = ChessLET.DEFAULT_LOCALE;
                break;

            case ID_EN: locale = "";  break;
            case ID_DE: locale = "de";  break;
            case ID_FR: locale = "fr";  break;
        }

        if (id != Selector.BACK)
            localizer.setLocale (locale);

        showMainMenu();
    }

    /**
     *
     * @param id
     */
    private void handleMainMenu (int id)
    {
        switch (id)
        {
            default:
                ChessLET.NotImplemented();
                return;

            case Selector.BACK:
                break;

            case ID_EXIT:
                ChessLET.Exit();
                return;

            case ID_TURN:
                turn = !turn;
                save (LOVER_ID);
                break;

            case ID_RESET:
                Message.Show (this, TAG_RESET, Message.CONFIRM, L._("T_ASK_BOARD_RESET"));
                return;
                
            case ID_UNDO_ALL:
                b.h.undoAll (b);
                initCursor();
                break;

            case ID_LANG:
                showLanguageMenu();
                return;

            case ID_POSITION:
                PositionCanvas.Show (true);
                return;

            case ID_BACKLIGHT:
                ChessCanvas.BacklightOn ^= true;
                break;

            case ID_FAMOUS:
//#if CONSOLE                
//#                 b.h.dump (History.DUMP_STRING, false);
//#endif
                showFamousMenu ();
                return;

            case ID_EXERCISE:
//#if CONSOLE   
//#                 b.h.undoAll (b);
//#                 b.dump();
//#                 initCursor();
//#endif
                showExerciseMenu();
                return;

            case ID_HELP:
                TextForm.Show (this, TAG_HELP, L._("T_TITLE"), getHelpItems ());
                return;
        }

        resumeGame();
    }

    private String[][] getHelpItems ()
    {
        return new String[][]
        {
            {L._("L_UDLR"), L._("PLAY.UDLR")}, 
            {L._("L_FIRE"), L._("PLAY.FIRE")}, 
            {"1-4", L._("PLAY.1-4")}, 
            {"1",   L._("PLAY.1")}, 
            {"2",   L._("PLAY.2")}, 
            {"3",   L._("PLAY.3")}, 
            {"4",   L._("PLAY.4")}, 
            {"7",   L._("PLAY.7")}, 
            {"8",   L._("PLAY.8")}, 
            {"9",   L._("PLAY.9")}, 
            {"*",   L._("PLAY.*")}, 
            {"#",   L._("PLAY.#")}
        };
    }
    
    public void messageReturned (int id, int tag)
    {
        if (TAG_HELP == tag)
        {
            showMainMenu();
        }
        
        if (TAG_RESET == tag)
        {
            if (Command.OK == id)
            {
                resetBoard ();
                resumeGame ();
            }
            if (Command.CANCEL == id)
            {
                showMainMenu();//resumeGame ();
            }
        }
    }

    protected static SelectorLabel GetLightLabel()
    {
        return lightLabel = new SelectorLabel (ID_BACKLIGHT,
                L._(BacklightOn
                    ? "T_BACKLIGHT_OFF"
                    : "T_BACKLIGHT_ON"));
    }

    public void languageChanged (String locale)
    {
        mainLabels = new SelectorLabel[]
        {
            new SelectorLabel (ID_TURN,      L._("T_TURN_BOARD")),
            new SelectorLabel (ID_POSITION,  L._("T_SAMPLE_POSITION")),
            new SelectorLabel (ID_RESET,     L._("T_RESET_BOARD")),
            new SelectorLabel (ID_UNDO_ALL,  L._("T_UNDO_ALL")),
            //new SelectorLabel (ID_LOAD,   L._("T_LOAD")),//     + L._("T_DOTS")),
            //new SelectorLabel (ID_SAVE,   L._("T_SAVE_AS")),//  + L._("T_DOTS")),
            new SelectorLabel (ID_FAMOUS,    L._("T_FAMOUS_GAMES")),
            new SelectorLabel (ID_EXERCISE,  L._("T_EXERCISES")),
            new SelectorLabel (ID_LANG,      L._("T_LANGUAGE")),// + L._("T_DOTS")),
            GetLightLabel(),
            new SelectorLabel (ID_HELP,      L._("T_HELP")),
            new SelectorLabel (ID_EXIT,      L._("T_EXIT"))
        };

        langLabels = new SelectorLabel[]
        {
            new SelectorLabel (ID_LANG_DEFAULT, L._("T_DEFAULT")),
            new SelectorLabel (ID_DE,  "Deutsch"),
            new SelectorLabel (ID_EN,  "English"),
            new SelectorLabel (ID_FR,  "Franais") // 
        };
    }
}
